@@ -2,11 +2,14 @@
 
 # pylint: disable=too-many-statements,line-too-long,invalid-unary-operand-type,too-many-lines
 import datetime
+import functools
 import hashlib
 import json
 import os
+import pickle
 
 import numpy as np
+import optuna
 import pandas as pd
 import pytz
 import wavetrainer as wt  # type: ignore
@@ -22,8 +25,6 @@ from sportsball.data.news_model import (NEWS_PUBLISHED_COLUMN,
 from sportsball.data.odds_model import (DT_COLUMN, ODDS_BET_COLUMN,
                                         ODDS_BOOKIE_COLUMN,
                                         ODDS_CANONICAL_COLUMN)
-from sportsball.data.player_model import \
-    ASSISTS_COLUMN as PLAYER_ASSISTS_COLUMN  # type: ignore
 from sportsball.data.player_model import \
     FIELD_GOALS_ATTEMPTED_COLUMN as \
     PLAYER_FIELD_GOALS_ATTEMPTED_COLUMN  # type: ignore
@@ -32,32 +33,45 @@ from sportsball.data.player_model import \
 from sportsball.data.player_model import \
     OFFENSIVE_REBOUNDS_COLUMN as PLAYER_OFFENSIVE_REBOUNDS_COLUMN
 from sportsball.data.player_model import (
-    PLAYER_ACLI_COLUMN, PLAYER_ASSIST_TACKLES_COLUMN, PLAYER_AT_BATS_COLUMN,
-    PLAYER_AVERAGE_GAIN_COLUMN, PLAYER_AVERAGE_INTERCEPTION_YARDS_COLUMN,
+    PLAYER_ACLI_COLUMN, PLAYER_AERIALS_LOST_COLUMN, PLAYER_AERIALS_WON_COLUMN,
+    PLAYER_ASSIST_TACKLES_COLUMN, PLAYER_ASSISTS_COLUMN, PLAYER_AT_BATS_COLUMN,
+    PLAYER_AVERAGE_DISTANCE_OF_DEFENSIVE_ACTIONS_COLUMN,
+    PLAYER_AVERAGE_GAIN_COLUMN, PLAYER_AVERAGE_GOAL_KICK_LENGTH_COLUMN,
+    PLAYER_AVERAGE_INTERCEPTION_YARDS_COLUMN,
     PLAYER_AVERAGE_KICKOFF_RETURN_YARDS_COLUMN,
     PLAYER_AVERAGE_KICKOFF_YARDS_COLUMN, PLAYER_AVERAGE_LEVERAGE_INDEX_COLUMN,
-    PLAYER_AVERAGE_PUNT_RETURN_YARDS_COLUMN, PLAYER_AVERAGE_SACK_YARDS_COLUMN,
-    PLAYER_AVERAGE_STUFF_YARDS_COLUMN, PLAYER_BASES_ON_BALLS_COLUMN,
+    PLAYER_AVERAGE_PASS_LENGTH_COLUMN, PLAYER_AVERAGE_PUNT_RETURN_YARDS_COLUMN,
+    PLAYER_AVERAGE_SACK_YARDS_COLUMN, PLAYER_AVERAGE_STUFF_YARDS_COLUMN,
+    PLAYER_BALL_RECOVERIES_COLUMN, PLAYER_BASES_ON_BALLS_COLUMN,
     PLAYER_BATTERS_FACED_COLUMN, PLAYER_BEHINDS_COLUMN,
     PLAYER_BIRTH_DATE_COLUMN, PLAYER_BLOCKED_FIELD_GOAL_TOUCHDOWNS_COLUMN,
     PLAYER_BLOCKED_PUNT_EZ_REC_TD_COLUMN,
     PLAYER_BLOCKED_PUNT_TOUCHDOWNS_COLUMN, PLAYER_BLOCKS_COLUMN,
-    PLAYER_BOUNCES_COLUMN, PLAYER_BROWNLOW_VOTES_COLUMN,
+    PLAYER_BOUNCES_COLUMN, PLAYER_BROWNLOW_VOTES_COLUMN, PLAYER_CARRIES_COLUMN,
+    PLAYER_CARRIES_INTO_FINAL_THIRD_COLUMN,
+    PLAYER_CARRIES_INTO_PENALTY_AREA_COLUMN, PLAYER_CHALLENGES_LOST_COLUMN,
     PLAYER_CLANGERS_COLUMN, PLAYER_CLEARANCES_COLUMN,
     PLAYER_COMPLETION_PERCENTAGE_COLUMN, PLAYER_COMPLETIONS_COLUMN,
     PLAYER_CONTESTED_MARKS_COLUMN, PLAYER_CONTESTED_POSSESSIONS_COLUMN,
-    PLAYER_CORSI_FOR_PERCENTAGE_COLUMN, PLAYER_CWPA_COLUMN,
-    PLAYER_DECISION_COLUMN, PLAYER_DEFENSIVE_FUMBLE_RETURN_YARDS_COLUMN,
+    PLAYER_CORNER_KICKS_COLUMN, PLAYER_CORSI_FOR_PERCENTAGE_COLUMN,
+    PLAYER_CROSSES_COLUMN, PLAYER_CROSSES_FACED_COLUMN,
+    PLAYER_CROSSES_INTO_PENALTY_AREA_COLUMN, PLAYER_CROSSES_STOPPED_COLUMN,
+    PLAYER_CWPA_COLUMN, PLAYER_DEAD_BALL_PASSES_COLUMN, PLAYER_DECISION_COLUMN,
+    PLAYER_DEFENSIVE_ACTIONS_OUTSIDE_PENALTY_AREA_COLUMN,
+    PLAYER_DEFENSIVE_FUMBLE_RETURN_YARDS_COLUMN,
     PLAYER_DEFENSIVE_FUMBLE_RETURNS_COLUMN,
     PLAYER_DEFENSIVE_FUMBLES_TOUCHDOWNS_COLUMN, PLAYER_DEFENSIVE_POINTS_COLUMN,
     PLAYER_DEFENSIVE_REBOUNDS_COLUMN, PLAYER_DEFENSIVE_TOUCHDOWNS_COLUMN,
     PLAYER_DEFENSIVE_ZONE_STARTS_COLUMN, PLAYER_DISPOSALS_COLUMN,
-    PLAYER_EARNED_RUNS_COLUMN, PLAYER_EFFECTIVE_FIELD_GOAL_PERCENTAGE_COLUMN,
-    PLAYER_ERA_COLUMN, PLAYER_ESPN_QUARTERBACK_RATING_COLUMN,
+    PLAYER_DISPOSSESSED_COLUMN, PLAYER_DRIBBLERS_TACKLED_COLUMN,
+    PLAYER_DRIBBLES_CHALLENGED_COLUMN, PLAYER_EARNED_RUNS_COLUMN,
+    PLAYER_EFFECTIVE_FIELD_GOAL_PERCENTAGE_COLUMN, PLAYER_ERA_COLUMN,
+    PLAYER_ERRORS_COLUMN, PLAYER_ESPN_QUARTERBACK_RATING_COLUMN,
     PLAYER_ESPN_RUNNINGBACK_RATING_COLUMN, PLAYER_ESPN_WIDERECEIVER_COLUMN,
     PLAYER_EVEN_STRENGTH_ASSISTS_COLUMN, PLAYER_EVEN_STRENGTH_GOALS_COLUMN,
-    PLAYER_EXTRA_POINT_ATTEMPTS_COLUMN, PLAYER_EXTRA_POINT_BLOCKED_COLUMN,
-    PLAYER_EXTRA_POINT_PERCENTAGE_COLUMN,
+    PLAYER_EXPECTED_ASSISTED_GOALS_COLUMN, PLAYER_EXPECTED_ASSISTS_COLUMN,
+    PLAYER_EXPECTED_GOALS_COLUMN, PLAYER_EXTRA_POINT_ATTEMPTS_COLUMN,
+    PLAYER_EXTRA_POINT_BLOCKED_COLUMN, PLAYER_EXTRA_POINT_PERCENTAGE_COLUMN,
     PLAYER_EXTRA_POINTS_BLOCKED_PERCENTAGE_COLUMN,
     PLAYER_EXTRA_POINTS_MADE_COLUMN, PLAYER_FAIR_CATCH_PERCENTAGE_COLUMN,
     PLAYER_FAIR_CATCHES_COLUMN, PLAYER_FIELD_GOAL_ATTEMPT_YARDS_COLUMN,
@@ -81,7 +95,8 @@ from sportsball.data.player_model import (
     PLAYER_FIELD_GOALS_MADE_YARDS_COLUMN,
     PLAYER_FIELD_GOALS_MISSED_YARDS_COLUMN,
     PLAYER_FIELD_GOALS_PERCENTAGE_COLUMN, PLAYER_FLY_BALLS_COLUMN,
-    PLAYER_FORCED_FUMBLES_COLUMN, PLAYER_FREE_KICKS_AGAINST_COLUMN,
+    PLAYER_FORCED_FUMBLES_COLUMN, PLAYER_FOULS_COMMITTED_COLUMN,
+    PLAYER_FOULS_DRAWN_COLUMN, PLAYER_FREE_KICKS_AGAINST_COLUMN,
     PLAYER_FREE_KICKS_FOR_COLUMN, PLAYER_FREE_THROWS_ATTEMPTED_COLUMN,
     PLAYER_FREE_THROWS_COLUMN, PLAYER_FREE_THROWS_PERCENTAGE_COLUMN,
     PLAYER_FUMBLE_RECOVERIES_COLUMN, PLAYER_FUMBLE_RECOVERY_YARDS_COLUMN,
@@ -89,6 +104,7 @@ from sportsball.data.player_model import (
     PLAYER_FUMBLES_RECOVERED_COLUMN, PLAYER_FUMBLES_RECOVERED_YARDS_COLUMN,
     PLAYER_FUMBLES_TOUCHDOWNS_COLUMN, PLAYER_GAME_SCORE_COLUMN,
     PLAYER_GAME_WINNING_GOALS_COLUMN, PLAYER_GOAL_ASSISTS_COLUMN,
+    PLAYER_GOAL_CREATING_ACTIONS_COLUMN, PLAYER_GOAL_KICKS_ATTEMPTED_COLUMN,
     PLAYER_GOALS_AGAINST_COLUMN, PLAYER_GOALS_COLUMN,
     PLAYER_GROSS_AVERAGE_PUNT_YARDS_COLUMN, PLAYER_GROUND_BALLS_COLUMN,
     PLAYER_HANDBALLS_COLUMN, PLAYER_HEADSHOT_COLUMN, PLAYER_HEIGHT_COLUMN,
@@ -96,9 +112,10 @@ from sportsball.data.player_model import (
     PLAYER_HOME_RUNS_COLUMN, PLAYER_HURRIES_COLUMN,
     PLAYER_INDIVIDUAL_CORSI_FOR_EVENTS_COLUMN, PLAYER_INHERITED_RUNNERS_COLUMN,
     PLAYER_INHERITED_SCORES_COLUMN, PLAYER_INNINGS_PITCHED_COLUMN,
-    PLAYER_INSIDES_COLUMN, PLAYER_INTERCEPTION_PERCENTAGE_COLUMN,
+    PLAYER_INSIDES_COLUMN, PLAYER_INSWINGING_CORNER_KICKS_COLUMN,
+    PLAYER_INTERCEPTION_PERCENTAGE_COLUMN,
     PLAYER_INTERCEPTION_TOUCHDOWNS_COLUMN, PLAYER_INTERCEPTION_YARDS_COLUMN,
-    PLAYER_INTERCEPTIONS_COLUMN,
+    PLAYER_INTERCEPTIONS_COLUMN, PLAYER_KEY_PASSES_COLUMN,
     PLAYER_KICK_RETURN_FAIR_CATCH_PERCENTAGE_COLUMN,
     PLAYER_KICK_RETURN_FAIR_CATCHES_COLUMN, PLAYER_KICK_RETURN_FUMBLES_COLUMN,
     PLAYER_KICK_RETURN_FUMBLES_LOST_COLUMN,
@@ -107,7 +124,8 @@ from sportsball.data.player_model import (
     PLAYER_KICKOFF_RETURN_YARDS_COLUMN, PLAYER_KICKOFF_RETURNS_COLUMN,
     PLAYER_KICKOFF_RETURNS_TOUCHDOWNS_COLUMN, PLAYER_KICKOFF_YARDS_COLUMN,
     PLAYER_KICKOFFS_COLUMN, PLAYER_KICKS_BLOCKED_COLUMN, PLAYER_KICKS_COLUMN,
-    PLAYER_LINE_DRIVES_COLUMN, PLAYER_LONG_FIELD_GOAL_ATTEMPT_COLUMN,
+    PLAYER_LINE_DRIVES_COLUMN, PLAYER_LIVE_BALL_PASSES_COLUMN,
+    PLAYER_LIVE_BALL_TOUCHES_COLUMN, PLAYER_LONG_FIELD_GOAL_ATTEMPT_COLUMN,
     PLAYER_LONG_FIELD_GOAL_MADE_COLUMN, PLAYER_LONG_INTERCEPTION_COLUMN,
     PLAYER_LONG_KICK_RETURN_COLUMN, PLAYER_LONG_KICKOFF_COLUMN,
     PLAYER_LONG_PASSING_COLUMN, PLAYER_LONG_PUNT_COLUMN,
@@ -116,14 +134,15 @@ from sportsball.data.player_model import (
     PLAYER_MARKS_INSIDE_COLUMN, PLAYER_MISC_FUMBLE_RETURN_YARDS_COLUMN,
     PLAYER_MISC_FUMBLE_RETURNS_COLUMN, PLAYER_MISC_POINTS_COLUMN,
     PLAYER_MISC_TOUCHDOWNS_COLUMN, PLAYER_MISC_YARDS_COLUMN,
-    PLAYER_MISSED_FIELD_GOAL_RETURN_TD_COLUMN,
+    PLAYER_MISCONTROLS_COLUMN, PLAYER_MISSED_FIELD_GOAL_RETURN_TD_COLUMN,
     PLAYER_NET_AVERAGE_PUNT_YARDS_COLUMN, PLAYER_NET_PASSING_ATTEMPTS_COLUMN,
     PLAYER_NET_PASSING_YARDS_COLUMN, PLAYER_NET_TOTAL_YARDS_COLUMN,
-    PLAYER_NET_YARDS_PER_PASS_ATTEMPT_COLUMN, PLAYER_OBP_COLUMN,
+    PLAYER_NET_YARDS_PER_PASS_ATTEMPT_COLUMN,
+    PLAYER_NON_PENALTY_EXPECTED_GOALS_COLUMN, PLAYER_OBP_COLUMN,
     PLAYER_OFFENSIVE_FUMBLES_TOUCHDOWNS_COLUMN,
     PLAYER_OFFENSIVE_TWO_POINT_RETURNS_COLUMN,
     PLAYER_OFFENSIVE_ZONE_START_PERCENTAGE_COLUMN,
-    PLAYER_OFFENSIVE_ZONE_STARTS_COLUMN,
+    PLAYER_OFFENSIVE_ZONE_STARTS_COLUMN, PLAYER_OFFSIDES_COLUMN,
     PLAYER_ON_SHOT_ICE_AGAINST_EVENTS_COLUMN,
     PLAYER_ON_SHOT_ICE_FOR_EVENTS_COLUMN, PLAYER_ONE_PERCENTERS_COLUMN,
     PLAYER_ONE_POINT_SAFETIES_MADE_COLUMN,
@@ -131,10 +150,22 @@ from sportsball.data.player_model import (
     PLAYER_OPPOSITION_FUMBLE_RECOVERY_YARDS_COLUMN,
     PLAYER_OPPOSITION_SPECIAL_TEAM_FUMBLE_RETURN_YARDS_COLUMN,
     PLAYER_OPPOSITION_SPECIAL_TEAM_FUMBLE_RETURNS_COLUMN, PLAYER_OPS_COLUMN,
-    PLAYER_PASSES_BATTED_DOWN_COLUMN, PLAYER_PASSES_DEFENDED_COLUMN,
-    PLAYER_PASSING_ATTEMPTS_COLUMN, PLAYER_PASSING_BIG_PLAYS_COLUMN,
-    PLAYER_PASSING_FIRST_DOWNS_COLUMN, PLAYER_PASSING_FUMBLES_COLUMN,
-    PLAYER_PASSING_FUMBLES_LOST_COLUMN,
+    PLAYER_OUTSWINGING_CORNER_KICKS_COLUMN, PLAYER_OWN_GOALS_COLUMN,
+    PLAYER_PASS_COMPLETION_COLUMN, PLAYER_PASS_COMPLETION_LONG_COLUMN,
+    PLAYER_PASS_COMPLETION_MEDIUM_COLUMN, PLAYER_PASS_COMPLETION_SHORT_COLUMN,
+    PLAYER_PASSES_ATTEMPTED_COLUMN, PLAYER_PASSES_ATTEMPTED_LONG_COLUMN,
+    PLAYER_PASSES_ATTEMPTED_MEDIUM_COLUMN,
+    PLAYER_PASSES_ATTEMPTED_MINUS_GOAL_KICKS_COLUMN,
+    PLAYER_PASSES_ATTEMPTED_SHORT_COLUMN, PLAYER_PASSES_BATTED_DOWN_COLUMN,
+    PLAYER_PASSES_BLOCKED_COLUMN, PLAYER_PASSES_COMPLETED_COLUMN,
+    PLAYER_PASSES_COMPLETED_LONG_COLUMN, PLAYER_PASSES_COMPLETED_MEDIUM_COLUMN,
+    PLAYER_PASSES_COMPLETED_SHORT_COLUMN, PLAYER_PASSES_DEFENDED_COLUMN,
+    PLAYER_PASSES_FROM_FREE_KICKS_COLUMN,
+    PLAYER_PASSES_INTO_FINAL_THIRD_COLUMN,
+    PLAYER_PASSES_INTO_PENALTY_AREA_COLUMN, PLAYER_PASSES_OFFSIDE_COLUMN,
+    PLAYER_PASSES_RECEIVED_COLUMN, PLAYER_PASSING_ATTEMPTS_COLUMN,
+    PLAYER_PASSING_BIG_PLAYS_COLUMN, PLAYER_PASSING_FIRST_DOWNS_COLUMN,
+    PLAYER_PASSING_FUMBLES_COLUMN, PLAYER_PASSING_FUMBLES_LOST_COLUMN,
     PLAYER_PASSING_TOUCHDOWN_PERCENTAGE_COLUMN,
     PLAYER_PASSING_TOUCHDOWNS_9_YARDS_COLUMN,
     PLAYER_PASSING_TOUCHDOWNS_19_YARDS_COLUMN,
@@ -144,11 +175,24 @@ from sportsball.data.player_model import (
     PLAYER_PASSING_TOUCHDOWNS_ABOVE_50_YARDS_COLUMN,
     PLAYER_PASSING_TOUCHDOWNS_COLUMN, PLAYER_PASSING_YARDS_AFTER_CATCH_COLUMN,
     PLAYER_PASSING_YARDS_AT_CATCH_COLUMN, PLAYER_PASSING_YARDS_COLUMN,
-    PLAYER_PENALTIES_IN_MINUTES_COLUMN, PLAYER_PERCENTAGE_PLAYED_COLUMN,
-    PLAYER_PERSONAL_FOULS_COLUMN, PLAYER_PITCHES_COLUMN,
-    PLAYER_PLATE_APPEARANCES_COLUMN, PLAYER_POINT_DIFFERENTIAL_COLUMN,
-    PLAYER_POINTS_ALLOWED_COLUMN, PLAYER_POINTS_COLUMN,
+    PLAYER_PENALTIES_IN_MINUTES_COLUMN, PLAYER_PENALTY_KICKS_ATTEMPTED_COLUMN,
+    PLAYER_PENALTY_KICKS_CONCEDED_COLUMN, PLAYER_PENALTY_KICKS_MADE_COLUMN,
+    PLAYER_PENALTY_KICKS_WON_COLUMN,
+    PLAYER_PERCENT_OF_DRIBBLERS_TACKLED_COLUMN,
+    PLAYER_PERCENTAGE_CROSSES_STOPPED_COLUMN,
+    PLAYER_PERCENTAGE_OF_AERIALS_WON_COLUMN,
+    PLAYER_PERCENTAGE_OF_GOAL_KICKS_THAT_WERE_LAUNCHED_COLUMN,
+    PLAYER_PERCENTAGE_OF_PASSES_THAT_WERE_LAUNCHED_COLUMN,
+    PLAYER_PERCENTAGE_PLAYED_COLUMN, PLAYER_PERSONAL_FOULS_COLUMN,
+    PLAYER_PITCHES_COLUMN, PLAYER_PLATE_APPEARANCES_COLUMN,
+    PLAYER_POINT_DIFFERENTIAL_COLUMN, PLAYER_POINTS_ALLOWED_COLUMN,
+    PLAYER_POINTS_COLUMN, PLAYER_POST_SHOT_EXPECTED_GOALS_COLUMN,
     PLAYER_POWER_PLAY_ASSISTS_COLUMN, PLAYER_POWER_PLAY_GOALS_COLUMN,
+    PLAYER_PROGRESSIVE_CARRIES_COLUMN,
+    PLAYER_PROGRESSIVE_CARRYING_DISTANCE_COLUMN,
+    PLAYER_PROGRESSIVE_PASSES_COLUMN,
+    PLAYER_PROGRESSIVE_PASSES_RECEIVED_COLUMN,
+    PLAYER_PROGRESSIVE_PASSING_DISTANCE_COLUMN,
     PLAYER_PUNT_RETURN_FAIR_CATCH_PERCENTAGE_COLUMN,
     PLAYER_PUNT_RETURN_FAIR_CATCHES_COLUMN, PLAYER_PUNT_RETURN_FUMBLES_COLUMN,
     PLAYER_PUNT_RETURN_FUMBLES_LOST_COLUMN,
@@ -174,7 +218,8 @@ from sportsball.data.player_model import (
     PLAYER_RECEIVING_TOUCHDOWNS_COLUMN,
     PLAYER_RECEIVING_YARDS_AFTER_CATCH_COLUMN,
     PLAYER_RECEIVING_YARDS_AT_CATCH_COLUMN, PLAYER_RECEIVING_YARDS_COLUMN,
-    PLAYER_RECEPTIONS_COLUMN, PLAYER_RELATIVE_CORSI_FOR_PERCENTAGE_COLUMN,
+    PLAYER_RECEPTIONS_COLUMN, PLAYER_RED_CARDS_COLUMN,
+    PLAYER_RELATIVE_CORSI_FOR_PERCENTAGE_COLUMN,
     PLAYER_RETURN_TOUCHDOWNS_COLUMN, PLAYER_RUNS_BATTED_IN_COLUMN,
     PLAYER_RUNS_SCORED_COLUMN, PLAYER_RUSHING_ATTEMPTS_COLUMN,
     PLAYER_RUSHING_BIG_PLAYS_COLUMN, PLAYER_RUSHING_FIRST_DOWNS_COLUMN,
@@ -190,26 +235,45 @@ from sportsball.data.player_model import (
     PLAYER_SACKS_UNASSISTED_COLUMN, PLAYER_SACKS_YARDS_COLUMN,
     PLAYER_SACKS_YARDS_LOST_COLUMN, PLAYER_SAFETIES_COLUMN,
     PLAYER_SAVE_PERCENTAGE_COLUMN, PLAYER_SAVES_COLUMN,
-    PLAYER_SECONDS_PLAYED_COLUMN, PLAYER_SHIFTS_COLUMN,
-    PLAYER_SHOOTING_PERCENTAGE_COLUMN, PLAYER_SHORT_HANDED_ASSISTS_COLUMN,
-    PLAYER_SHORT_HANDED_GOALS_COLUMN, PLAYER_SHOTS_AGAINST_COLUMN,
-    PLAYER_SHOTS_ON_GOAL_COLUMN, PLAYER_SHUTOUTS_COLUMN, PLAYER_SLG_COLUMN,
+    PLAYER_SECOND_YELLOW_CARD_COLUMN, PLAYER_SECONDS_PLAYED_COLUMN,
+    PLAYER_SHIFTS_COLUMN, PLAYER_SHOOTING_PERCENTAGE_COLUMN,
+    PLAYER_SHORT_HANDED_ASSISTS_COLUMN, PLAYER_SHORT_HANDED_GOALS_COLUMN,
+    PLAYER_SHOT_CREATING_ACTIONS_COLUMN, PLAYER_SHOTS_AGAINST_COLUMN,
+    PLAYER_SHOTS_BLOCKED_COLUMN, PLAYER_SHOTS_ON_GOAL_COLUMN,
+    PLAYER_SHOTS_ON_TARGET_AGAINST_COLUMN, PLAYER_SHOTS_ON_TARGET_COLUMN,
+    PLAYER_SHOTS_TOTAL_COLUMN, PLAYER_SHUTOUTS_COLUMN, PLAYER_SLG_COLUMN,
     PLAYER_SOLO_TACKLES_COLUMN, PLAYER_SPECIAL_TEAM_FUMBLE_RETURN_YARDS_COLUMN,
     PLAYER_SPECIAL_TEAM_FUMBLE_RETURNS_COLUMN, PLAYER_STEALS_COLUMN,
-    PLAYER_STRIKEOUTS_COLUMN, PLAYER_STRIKES_BY_CONTACT_COLUMN,
-    PLAYER_STRIKES_COLUMN, PLAYER_STRIKES_LOOKING_COLUMN,
-    PLAYER_STRIKES_SWINGING_COLUMN, PLAYER_STUFF_YARDS_COLUMN,
-    PLAYER_STUFF_YARDS_LOST, PLAYER_STUFFS_COLUMN, PLAYER_TACKLES_COLUMN,
-    PLAYER_TACKLES_FOR_LOSS_COLUMN, PLAYER_TACKLES_YARDS_LOST_COLUMN,
+    PLAYER_STRAIGHT_CORNER_KICKS_COLUMN, PLAYER_STRIKEOUTS_COLUMN,
+    PLAYER_STRIKES_BY_CONTACT_COLUMN, PLAYER_STRIKES_COLUMN,
+    PLAYER_STRIKES_LOOKING_COLUMN, PLAYER_STRIKES_SWINGING_COLUMN,
+    PLAYER_STUFF_YARDS_COLUMN, PLAYER_STUFF_YARDS_LOST, PLAYER_STUFFS_COLUMN,
+    PLAYER_SUCCESSFUL_TAKE_ON_PERCENTAGE_COLUMN,
+    PLAYER_SUCCESSFUL_TAKE_ONS_COLUMN, PLAYER_SWITCHES_COLUNM,
+    PLAYER_TACKLED_DURING_TAKE_ON_PERCENTAGE_COLUMN, PLAYER_TACKLES_COLUMN,
+    PLAYER_TACKLES_FOR_LOSS_COLUMN, PLAYER_TACKLES_IN_ATTACKING_THIRD_COLUMN,
+    PLAYER_TACKLES_IN_DEFENSIVE_THIRD_COLUMN,
+    PLAYER_TACKLES_IN_MIDDLE_THIRD_COLUMN,
+    PLAYER_TACKLES_PLUS_INTERCEPTIONS_COLUMN, PLAYER_TACKLES_WON_COLUMN,
+    PLAYER_TACKLES_YARDS_LOST_COLUMN, PLAYER_TAKE_ONS_ATTEMPTED_COLUMN,
+    PLAYER_THREE_POINT_ATTEMPT_RATE_COLUMN,
     PLAYER_THREE_POINT_FIELD_GOALS_ATTEMPTED_COLUMN,
     PLAYER_THREE_POINT_FIELD_GOALS_COLUMN,
     PLAYER_THREE_POINT_FIELD_GOALS_PERCENTAGE_COLUMN,
-    PLAYER_TIME_ON_ICE_COLUMN, PLAYER_TOTAL_KICKING_POINTS_COLUMN,
-    PLAYER_TOTAL_OFFENSIVE_PLAYS_COLUMN, PLAYER_TOTAL_POINTS_COLUMN,
-    PLAYER_TOTAL_REBOUNDS_COLUMN, PLAYER_TOTAL_TOUCHDOWNS_COLUMN,
-    PLAYER_TOTAL_TWO_POINT_CONVERSIONS_COLUMN, PLAYER_TOTAL_YARDS_COLUMN,
-    PLAYER_TOTAL_YARDS_FROM_SCRIMMAGE_COLUMN,
+    PLAYER_THROUGH_BALLS_COLUMN, PLAYER_THROW_INS_TAKEN_COLUMN,
+    PLAYER_THROWS_ATTEMPTED_COLUMN, PLAYER_TIME_ON_ICE_COLUMN,
+    PLAYER_TIMES_TACKLED_DURING_TAKE_ONS_COLUMN,
+    PLAYER_TOTAL_CARRYING_DISTANCE_COLUMN, PLAYER_TOTAL_KICKING_POINTS_COLUMN,
+    PLAYER_TOTAL_OFFENSIVE_PLAYS_COLUMN, PLAYER_TOTAL_PASSING_DISTANCE_COLUMN,
+    PLAYER_TOTAL_POINTS_COLUMN, PLAYER_TOTAL_REBOUNDS_COLUMN,
+    PLAYER_TOTAL_TOUCHDOWNS_COLUMN, PLAYER_TOTAL_TWO_POINT_CONVERSIONS_COLUMN,
+    PLAYER_TOTAL_YARDS_COLUMN, PLAYER_TOTAL_YARDS_FROM_SCRIMMAGE_COLUMN,
     PLAYER_TOUCHBACK_PERCENTAGE_COLUMN, PLAYER_TOUCHBACKS_COLUMN,
+    PLAYER_TOUCHES_COLUMN, PLAYER_TOUCHES_IN_ATTACKING_PENALTY_AREA_COLUMN,
+    PLAYER_TOUCHES_IN_ATTACKING_THIRD_COLUMN,
+    PLAYER_TOUCHES_IN_DEFENSIVE_PENALTY_AREA_COLUMN,
+    PLAYER_TOUCHES_IN_DEFENSIVE_THIRD_COLUMN,
+    PLAYER_TOUCHES_IN_MIDDLE_THIRD_COLUMN,
     PLAYER_TRUE_SHOOTING_PERCENTAGE_COLUMN,
     PLAYER_TWO_POINT_PASS_ATTEMPT_COLUMN, PLAYER_TWO_POINT_PASS_COLUMN,
     PLAYER_TWO_POINT_RECEPTION_ATTEMPTS_COLUMN,
@@ -219,7 +283,8 @@ from sportsball.data.player_model import (
     PLAYER_WPA_PLUS_COLUMN, PLAYER_YARDS_ALLOWED_COLUMN,
     PLAYER_YARDS_PER_COMPLETION_COLUMN, PLAYER_YARDS_PER_KICK_RETURN_COLUMN,
     PLAYER_YARDS_PER_PASS_ATTEMPT_COLUMN, PLAYER_YARDS_PER_RECEPTION_COLUMN,
-    PLAYER_YARDS_PER_RETURN_COLUMN, PLAYER_YARDS_PER_RUSH_ATTEMPT_COLUMN)
+    PLAYER_YARDS_PER_RETURN_COLUMN, PLAYER_YARDS_PER_RUSH_ATTEMPT_COLUMN,
+    PLAYER_YELLOW_CARDS_COLUMN)
 from sportsball.data.player_model import \
     TURNOVERS_COLUMN as PLAYER_TURNOVERS_COLUMN  # type: ignore
 from sportsball.data.team_model import ASSISTS_COLUMN  # type: ignore
@@ -270,6 +335,8 @@ _CONFIG_FILENAME = "config.json"
 _PLACE_KEY = "place"
 _VALIDATION_SIZE = datetime.timedelta(days=365)
 _TEST_SIZE = datetime.timedelta(days=365)
+_SAMPLER_FILENAME = "sampler.pkl"
+_KELLY_KEY = "kelly"
 
 
 class Strategy:
@@ -317,6 +384,20 @@ class Strategy:
 
         self._returns = None
 
+        storage_name = f"sqlite:///{name}/study.db"
+        sampler_file = os.path.join(name, _SAMPLER_FILENAME)
+        restored_sampler = None
+        if os.path.exists(sampler_file):
+            with open(sampler_file, "rb") as handle:
+                restored_sampler = pickle.load(handle)
+        self._study = optuna.create_study(
+            study_name=name,
+            storage=storage_name,
+            load_if_exists=True,
+            sampler=restored_sampler,
+            direction=optuna.study.StudyDirection.MAXIMIZE,
+        )
+
     @property
     def df(self) -> pd.DataFrame | None:
         """Fetch the dataframe currently being operated on."""
@@ -337,7 +418,7 @@ class Strategy:
         """Fetch the name of the strategy."""
         return self._name
 
-    def kelly_ratio(self, df: pd.DataFrame) -> float:
+    def find_returns(self, df: pd.DataFrame) -> pd.Series:
         """Find the best kelly ratio for this strategy."""
         main_df = self.df
         if main_df is None:
@@ -346,25 +427,40 @@ class Strategy:
         df[points_cols] = main_df[points_cols].to_numpy()
         cutoff_dt = pd.to_datetime(datetime.datetime.now() - _VALIDATION_SIZE).date()
         df = df[df[GAME_DT_COLUMN].dt.date > cutoff_dt]
-        df = augment_kelly_fractions(df, len(points_cols))
-        df.to_parquet(os.path.join(self._name, "returns_df.parquet.gzip"))
-        max_return = 0.0
-        max_kelly = 0.0
-        for i in range(100):
-            test_kelly_ratio = (100 - i) / 100.0
+
+        def trial_returns(
+            trial: optuna.Trial | optuna.trial.FrozenTrial, df: pd.DataFrame
+        ) -> pd.Series:
+            alpha = trial.suggest_float("alpha", 0.0, 2.0)
+            kelly_threshold = trial.suggest_float(_KELLY_KEY, 0.0, 1.0)
+
+            df = augment_kelly_fractions(df, len(points_cols), alpha)
             returns = calculate_returns(
-                test_kelly_ratio,
-                df.copy(),
+                kelly_threshold,
+                df,
                 self._name,
             )
-            value = calculate_value(returns)
-            if value > max_return or max_kelly == 0.0:
-                max_return = value
-                max_kelly = test_kelly_ratio
-        print(f"Max Kelly: {max_kelly}")
-        self._returns = calculate_returns(max_kelly, df.copy(), self._name)
+            return returns
 
-        return max_kelly
+        def run_trial(
+            trial: optuna.Trial | optuna.trial.FrozenTrial, df: pd.DataFrame
+        ) -> float:
+            returns = trial_returns(trial, df)
+            value = calculate_value(returns)
+            return value
+
+        if not self._study.trials:
+            self._study.optimize(
+                functools.partial(
+                    run_trial,
+                    df=df,
+                ),
+                n_trials=100,
+                timeout=60.0 * 60.0 * 5,
+                show_progress_bar=True,
+            )
+
+        return trial_returns(self._study.best_trial, df)
 
     def fit(self):
         """Fits the strategy to the dataset by walking forward."""
@@ -434,8 +530,7 @@ class Strategy:
     def returns(self) -> pd.Series:
         """Render the returns of the strategy."""
         df = self.predict()
-        self.kelly_ratio(df)
-        returns = self._returns
+        returns = self.find_returns(df)
         if returns is None:
             raise ValueError("returns is null")
         return returns
@@ -448,7 +543,8 @@ class Strategy:
         """Find the next predictions for betting."""
         dt_column = DELIMITER.join([GAME_DT_COLUMN])
         df = self.predict()
-        kelly_ratio = self.kelly_ratio(df)
+        self.find_returns(df)
+        kelly_ratio = self._study.best_trial.suggest_float(_KELLY_KEY, 0.0, 1.0)
         start_dt = datetime.datetime.now(datetime.timezone.utc)
         end_dt = start_dt + datetime.timedelta(days=3.0)
         df = df[df[dt_column] > start_dt]
@@ -936,6 +1032,109 @@ class Strategy:
                                 PLAYER_INHERITED_RUNNERS_COLUMN,
                                 PLAYER_INHERITED_SCORES_COLUMN,
                                 PLAYER_EFFECTIVE_FIELD_GOAL_PERCENTAGE_COLUMN,
+                                PLAYER_PENALTY_KICKS_MADE_COLUMN,
+                                PLAYER_PENALTY_KICKS_ATTEMPTED_COLUMN,
+                                PLAYER_SHOTS_TOTAL_COLUMN,
+                                PLAYER_SHOTS_ON_TARGET_COLUMN,
+                                PLAYER_YELLOW_CARDS_COLUMN,
+                                PLAYER_RED_CARDS_COLUMN,
+                                PLAYER_TOUCHES_COLUMN,
+                                PLAYER_EXPECTED_GOALS_COLUMN,
+                                PLAYER_NON_PENALTY_EXPECTED_GOALS_COLUMN,
+                                PLAYER_EXPECTED_ASSISTED_GOALS_COLUMN,
+                                PLAYER_SHOT_CREATING_ACTIONS_COLUMN,
+                                PLAYER_GOAL_CREATING_ACTIONS_COLUMN,
+                                PLAYER_PASSES_COMPLETED_COLUMN,
+                                PLAYER_PASSES_ATTEMPTED_COLUMN,
+                                PLAYER_PASS_COMPLETION_COLUMN,
+                                PLAYER_PROGRESSIVE_PASSES_COLUMN,
+                                PLAYER_CARRIES_COLUMN,
+                                PLAYER_PROGRESSIVE_CARRIES_COLUMN,
+                                PLAYER_TAKE_ONS_ATTEMPTED_COLUMN,
+                                PLAYER_SUCCESSFUL_TAKE_ONS_COLUMN,
+                                PLAYER_TOTAL_PASSING_DISTANCE_COLUMN,
+                                PLAYER_PROGRESSIVE_PASSING_DISTANCE_COLUMN,
+                                PLAYER_PASSES_COMPLETED_SHORT_COLUMN,
+                                PLAYER_PASSES_ATTEMPTED_SHORT_COLUMN,
+                                PLAYER_PASS_COMPLETION_SHORT_COLUMN,
+                                PLAYER_PASSES_COMPLETED_MEDIUM_COLUMN,
+                                PLAYER_PASSES_ATTEMPTED_MEDIUM_COLUMN,
+                                PLAYER_PASS_COMPLETION_MEDIUM_COLUMN,
+                                PLAYER_PASSES_COMPLETED_LONG_COLUMN,
+                                PLAYER_PASSES_ATTEMPTED_LONG_COLUMN,
+                                PLAYER_PASS_COMPLETION_LONG_COLUMN,
+                                PLAYER_EXPECTED_ASSISTS_COLUMN,
+                                PLAYER_KEY_PASSES_COLUMN,
+                                PLAYER_PASSES_INTO_FINAL_THIRD_COLUMN,
+                                PLAYER_PASSES_INTO_PENALTY_AREA_COLUMN,
+                                PLAYER_CROSSES_INTO_PENALTY_AREA_COLUMN,
+                                PLAYER_LIVE_BALL_PASSES_COLUMN,
+                                PLAYER_DEAD_BALL_PASSES_COLUMN,
+                                PLAYER_PASSES_FROM_FREE_KICKS_COLUMN,
+                                PLAYER_THROUGH_BALLS_COLUMN,
+                                PLAYER_SWITCHES_COLUNM,
+                                PLAYER_CROSSES_COLUMN,
+                                PLAYER_THROW_INS_TAKEN_COLUMN,
+                                PLAYER_CORNER_KICKS_COLUMN,
+                                PLAYER_INSWINGING_CORNER_KICKS_COLUMN,
+                                PLAYER_OUTSWINGING_CORNER_KICKS_COLUMN,
+                                PLAYER_STRAIGHT_CORNER_KICKS_COLUMN,
+                                PLAYER_PASSES_OFFSIDE_COLUMN,
+                                PLAYER_PASSES_BLOCKED_COLUMN,
+                                PLAYER_TACKLES_WON_COLUMN,
+                                PLAYER_TACKLES_IN_DEFENSIVE_THIRD_COLUMN,
+                                PLAYER_TACKLES_IN_MIDDLE_THIRD_COLUMN,
+                                PLAYER_TACKLES_IN_ATTACKING_THIRD_COLUMN,
+                                PLAYER_DRIBBLERS_TACKLED_COLUMN,
+                                PLAYER_DRIBBLES_CHALLENGED_COLUMN,
+                                PLAYER_PERCENT_OF_DRIBBLERS_TACKLED_COLUMN,
+                                PLAYER_CHALLENGES_LOST_COLUMN,
+                                PLAYER_SHOTS_BLOCKED_COLUMN,
+                                PLAYER_TACKLES_PLUS_INTERCEPTIONS_COLUMN,
+                                PLAYER_ERRORS_COLUMN,
+                                PLAYER_TOUCHES_IN_DEFENSIVE_PENALTY_AREA_COLUMN,
+                                PLAYER_TOUCHES_IN_DEFENSIVE_THIRD_COLUMN,
+                                PLAYER_TOUCHES_IN_MIDDLE_THIRD_COLUMN,
+                                PLAYER_TOUCHES_IN_ATTACKING_THIRD_COLUMN,
+                                PLAYER_TOUCHES_IN_ATTACKING_PENALTY_AREA_COLUMN,
+                                PLAYER_LIVE_BALL_TOUCHES_COLUMN,
+                                PLAYER_SUCCESSFUL_TAKE_ON_PERCENTAGE_COLUMN,
+                                PLAYER_TIMES_TACKLED_DURING_TAKE_ONS_COLUMN,
+                                PLAYER_TACKLED_DURING_TAKE_ON_PERCENTAGE_COLUMN,
+                                PLAYER_TOTAL_CARRYING_DISTANCE_COLUMN,
+                                PLAYER_PROGRESSIVE_CARRYING_DISTANCE_COLUMN,
+                                PLAYER_CARRIES_INTO_FINAL_THIRD_COLUMN,
+                                PLAYER_CARRIES_INTO_PENALTY_AREA_COLUMN,
+                                PLAYER_MISCONTROLS_COLUMN,
+                                PLAYER_DISPOSSESSED_COLUMN,
+                                PLAYER_PASSES_RECEIVED_COLUMN,
+                                PLAYER_PROGRESSIVE_PASSES_RECEIVED_COLUMN,
+                                PLAYER_SECOND_YELLOW_CARD_COLUMN,
+                                PLAYER_FOULS_COMMITTED_COLUMN,
+                                PLAYER_FOULS_DRAWN_COLUMN,
+                                PLAYER_OFFSIDES_COLUMN,
+                                PLAYER_PENALTY_KICKS_WON_COLUMN,
+                                PLAYER_PENALTY_KICKS_CONCEDED_COLUMN,
+                                PLAYER_OWN_GOALS_COLUMN,
+                                PLAYER_BALL_RECOVERIES_COLUMN,
+                                PLAYER_AERIALS_WON_COLUMN,
+                                PLAYER_AERIALS_LOST_COLUMN,
+                                PLAYER_PERCENTAGE_OF_AERIALS_WON_COLUMN,
+                                PLAYER_SHOTS_ON_TARGET_AGAINST_COLUMN,
+                                PLAYER_POST_SHOT_EXPECTED_GOALS_COLUMN,
+                                PLAYER_PASSES_ATTEMPTED_MINUS_GOAL_KICKS_COLUMN,
+                                PLAYER_THROWS_ATTEMPTED_COLUMN,
+                                PLAYER_PERCENTAGE_OF_PASSES_THAT_WERE_LAUNCHED_COLUMN,
+                                PLAYER_AVERAGE_PASS_LENGTH_COLUMN,
+                                PLAYER_GOAL_KICKS_ATTEMPTED_COLUMN,
+                                PLAYER_PERCENTAGE_OF_GOAL_KICKS_THAT_WERE_LAUNCHED_COLUMN,
+                                PLAYER_AVERAGE_GOAL_KICK_LENGTH_COLUMN,
+                                PLAYER_CROSSES_FACED_COLUMN,
+                                PLAYER_CROSSES_STOPPED_COLUMN,
+                                PLAYER_PERCENTAGE_CROSSES_STOPPED_COLUMN,
+                                PLAYER_DEFENSIVE_ACTIONS_OUTSIDE_PENALTY_AREA_COLUMN,
+                                PLAYER_AVERAGE_DISTANCE_OF_DEFENSIVE_ACTIONS_COLUMN,
+                                PLAYER_THREE_POINT_ATTEMPT_RATE_COLUMN,
                             ]
                         ],
                         player_column_prefix(i, x),
